from typing import Literal
from statsmodels.regression.linear_model import RegressionResultsWrapper
from estyp.testing.__base import TestResults, __var_test, __t_test, __nested_models_test
from pandas import Series

def var_test(
    x,
    y,
    ratio=1,
    alternative: Literal["two-sided", "less", "greater"] = "two-sided",
    conf_level=0.95,
) -> TestResults:
    """
# F Test to Compare Two Variances
## Description
Performs an F test to compare the variances of two samples from normal populations. This function is inspired by the `var.test()` function of the software R.

## Usage

```python
var_test(x, y, ...)
var_test(x, y, ratio = 1,
        alternative = [two-sided", "less", "greater"],
        conf_level = 0.95)
```

## Arguments
* `x`, `y`: numeric list, `np.array` or `pd.Series` of data values.
* `ratio`: the hypothesized ratio of the population variances of x and y.
* `alternative`: a character string specifying the alternative hypothesis, must be one of "two.sided" (default), "greater" or "less". You can specify just the initial letter.
* `conf_level`: a number between 0 and 1 indicanting the confidence level of the interval.


## Details
The null hypothesis is that the ratio of the variances of the populations from which x and y were drawn, is equal to ratio.

## Value
An instance of the `TestResults` class containing the following attributes:

* `statistic`: the value of the F test statistic.	
* `df`: the degrees of freedom for the F test statistic.
* `p_value`: the p-value for the test.
* `ci`: a confidence interval for the ratio of the population variances.
* `estimate`: the ratio of the sample variances of x and y.
* `alternative`: a string describing the alternative hypothesis.

## Examples

```python
import numpy as np
np.random.seed(2023)
x = np.random.normal(size=100)
y = np.random.normal(size=100)

print(var_test(x, y))
print(var_test(x, y, alternative="less"))
print(var_test(x, y, ratio = 0.9, alternative="greater"))
```
    """
    return __var_test(x, y, ratio, alternative, conf_level)


def t_test(
    x,
    y=None,
    alternative: Literal["two-sided", "less", "greater"] = "two-sided",
    mu:float    = 0,
    paired      = False,
    var_equal   = False,
    conf_level  = 0.95
):
    """

# Student's t-Test
## Description
Performs one and two sample t-tests on groups of data. This function is inspired by the `t.test()` function of the software R.

## Usage
```python
t_test(x, ...)
```

```python
t_test(x, y = None,
    alternative = ["two-sided", "less", "greater"],
    mu = 0, paired = False, var_equal = False,
    conf_level = 0.95)
```


## Arguments
* `x`: a (non-empty) numeric container of data values.
* `y`: an (optional) numeric container of data values.
* `alternative`: a string specifying the alternative hypothesis, must be one of "two-sided" (default), "greater" or "less".
* `mu`: a number indicating the true value of the mean (or difference in means if you are performing a two sample test).
* `paired`: a logical indicating whether you want a paired t-test.
* `var_equal`: a logical variable indicating whether to treat the two variances as being equal. If `True` then the pooled variance is used to estimate the variance otherwise the Welch (or Satterthwaite) approximation to the degrees of freedom is used.
* `conf_level`: a number between 0 and 1 indicanting the confidence level of the interval.


## Details
alternative = "greater" is the alternative that x has a larger mean than y. For the one-sample case: that the mean is positive.

If `paired` is `True` then both x and y must be specified and they must be the same length. Missing values are silently removed (in pairs if `paired` is `True`). If `var_equal` is `True` then the pooled estimate of the variance is used. By default, if `var_equal` is `False` then the variance is estimated separately for both groups and the Welch modification to the degrees of freedom is used.


### Value
An instance of the `TestResults` class containing the following attributes:

* `statistic`: the value of the t-statistic.	
* `df`: the degrees of freedom for the t-statistic.
* `p_value`: the p-value for the test.
* `ci`: a confidence interval for the mean appropriate to the specified alternative hypothesis.
* `estimate`: the estimated mean or list of estimated means depending on whether it was a one-sample test or a two-sample test.
* `alternative`: a character string describing the alternative hypothesis.
* `mu`: the mean of the null hypothesis.


## Examples
```python
import numpy as np
np.random.seed(2023)
x = np.random.normal(size=100)
y = np.random.normal(size=100)
mu = 0.1

print(t_test(x, mu=mu, alternative="less"))
print(t_test(x, y, mu=mu))
print(t_test(x, y, mu=mu, var_equal=True, alternative="greater"))
print(t_test(x, y, mu=mu, paired=True))
```
    """
    return __t_test(x, y, alternative, mu, paired, var_equal, conf_level)


def nested_models_test(
    fitted_small_model: RegressionResultsWrapper, 
    fitted_big_model: RegressionResultsWrapper
):
    """
    # Nested Models F-Test Function
    
    ## Description
    
    This function performs a nested models F-test using deviance from two fitted models from statsmodels library. The test compares two nested models: a larger or "big" model and a smaller or "small" model. The purpose of this test is to determine whether the larger model significantly improves the model fit compared to the smaller model by adding additional predictors.
    
    ## Parameters
    * `fitted_small_model`: The fitted model representing the smaller/nested model. It have to come from statsmodels.
    * `fitted_big_model`: : The fitted model representing the larger model, which includes all the predictors from the smaller model and potentially additional predictors. It have to come from statsmodels.
    
    ## Returns
    
    The function returns an object of class TestResults that contains the following information:

    * `method`: A string indicating the name of the statistical test (Nested models F-test).
    * `statistic`: The computed F-statistic value.
    * `estimate`: The difference in deviances between the models.
    * `df`: A dictionary with the degrees of freedom for the numerator and denominator of the F-statistic.
    * p_value: The p-value associated with the F-statistic.
    
    ## Examples
    ### Example 1: With OLS
    ```python
    import pandas as pd
    import statsmodels.api as sm
    data = pd.DataFrame({
        "x": [2.01, 2.99, 4.01, 5.01, 6.89],
        "y": [2, 3, 4, 5, 6]
    })
    
    model_small = sm.OLS.from_formula("y ~ 1", data).fit()
    model_big = sm.OLS.from_formula("y ~ x", data).fit()
    
    print(nested_models_test(model_small, model_big))
    ```
    ### Example 2: With Logit
    ```python
    import pandas as pd
    import statsmodels.api as sm
    data = pd.DataFrame({
        "x": [2.01, 2.99, 4.01, 3.01, 4.89],
        "y": [0, 1, 1, 0, 1]
    })
    
    model_small = sm.Logit.from_formula("y ~ 1", data).fit()
    model_big = sm.Logit.from_formula("y ~ x", data).fit()
    
    print(nested_models_test(model_small, model_big))
    ```
    ### Example 3: With GLM
    ```python
    import pandas as pd
    import statsmodels.api as sm
    data = pd.DataFrame({
        "x": [2.01, 2.99, 4.01, 5.01, 6.89],
        "y": [2, 3, 4, 5, 6]
    })
    model_small = sm.GLM.from_formula("y ~ 1", data, family = sm.families.Gamma()).fit()
    model_big = sm.GLM.from_formula("y ~ x", data, family = sm.families.Gamma()).fit()

    print(nested_models_test(model_small, model_big))
    ```
    """
    return __nested_models_test(fitted_small_model, fitted_big_model)


from scipy.stats import f as fisher
from estyp.testing.__base import __nested_models_test




def __both_selection(formula, data, model, max_iter, **kwargs) -> str:
    
    print("Este proceso tarda un buen tiempo ¡Paciencia! Momento de cuestionarte si eres feliz.")
    # Preparación
    fi = formula
    e = model.from_formula(fi, data, **kwargs)
    m = e.fit()
    preds = e.formula.split("+")
    vr = preds[0].split("~")[0].strip()
    preds[0] = preds[0].split("~")[1]
    preds = [x.strip() for x in preds]
    
    # Algoritmo
    d = data.copy()

    aicf = m.aic
    aic0 = 0

    var_el = []
    var_ag = []
    print(f"AIC Inicial: {aicf:0.2f}")

    for i in range(max_iter):
        if i == 0:
            aics = []
            for p in preds:
                # Crear formula
                p0 = d.drop(columns=[p, vr]).columns.tolist()
                f0 = f"{vr} ~ {' + '.join(p0) if p0 else 1}"
                # Ajustar modelo y extraer AIC
                m0 = model.from_formula(f0, d, **kwargs).fit()
                aics.append(m0.aic)
            aic0 = min(aics)
            id_min_aic = aics.index(min(aics))
            if aic0 < aicf:
                var_min = preds[id_min_aic]
                data0 = d.drop(columns=var_min)
                preds0 = data0.drop(columns=vr).columns.tolist()
                ff = f"{vr} ~ {' + '.join(preds0) if preds0 else 1}"
                var_el.append(var_min)
            else:
                print("No hay mejoras en el AIC, por lo que no se eliminarán variables.")
                return fi
        else:
            aicf = aic0
            # Quitar variable
            aics_a = []
            for p in preds0:
                p0 = data0.drop(columns=[p, vr]).columns.tolist()
                f0 = f"{vr} ~ {' + '.join(p0) if p0 else 1}"
                m0 = model.from_formula(f0, d, **kwargs).fit()
                aics_a.append(m0.aic)
            # Añadir variable
            aics_b = []
            for p in preds:
                f0 = f"{ff} {('+ ' + p) if p not in preds0 else ''}"
                m0 = model.from_formula(f0, d, **kwargs).fit() if p not in preds0 else None
                aics_b.append(m0.aic if p not in preds0 else aics_a[preds0.index(p)])
            min_a, min_b = min(aics_a), min(aics_b)
            aic0 = min_a if min_a < min_b else min_b
            if aic0 < aicf:
                if min_a < min_b:
                    var_min = preds0[aics_a.index(min_a)]
                    data0.drop(columns=var_min, inplace=True)
                    preds0 = data0.drop(columns=vr).columns.tolist()
                    ff = f"{vr} ~ {' + '.join(preds0) if preds0 else 1}"
                    var_el.append(var_min)
                else:
                    var_min = preds[aics_b.index(min_b)]
                    data0 = d[preds0 + [var_min, vr]]
                    preds0.append(var_min)
                    ff = f"{vr} ~ {' + '.join(preds0) if preds0 else 1}"
                    var_ag.append(var_min)
                aicf = aic0
            else:
                print("=========================")
                print("|| Fin de la selección ||")
                print("=========================")
                print(f"AIC Obtenido: {aicf:0.2f}")
                print("Variables eliminadas:", var_el if var_el else "Ninguna")
                print("Variables agregadas:", var_ag if var_ag else "Ninguna")
                print("Fórmula obtenida:", ff)
                return ff
        print(f"Iteración {i + 1} Finalizada | AIC Actual: {aic0:0.2f}")
        if not preds0:
            return f"{vr} + 1"
    print("Máximas iteraciones alcanzadas")
    return ff


def __forward_selection(y, data, model, alpha, **kwargs):
    preds = data.columns.to_list()
    preds.remove(y)

    f_actual = f"{y} ~ 1"
    m_actual = model.from_formula(f_actual, data, **kwargs).fit(disp=0)
    termino = False
    
    while not termino:
        valores_p = []
        for p in preds:
            f_prueba = f"{f_actual} + {p}"
            m_prueba = model.from_formula(f_prueba, data, **kwargs).fit(disp=0)
            pv = __nested_models_test(m_actual, m_prueba).p_value
            valores_p.append(pv)
        min_vp = min(valores_p)
        if min_vp >= alpha:
            termino = True
            f_actual = f_actual.replace(" 1 +", "")
            m_actual = model.from_formula(f_actual, data, **kwargs).fit(disp=0)
        else:
            var_min = preds[valores_p.index(min_vp)]
            f_actual = f"{f_actual} + {var_min}"
            m_actual = model.from_formula(f_actual, data, **kwargs).fit(disp=0)
            preds.remove(var_min)
            termino = False if preds else True
            vp = f"{min_vp:0.4f}" if min_vp >= 0.0001 else "<0.0001"
            print(f"Variable agregada: {var_min:30} | valor-p: {vp}")
    print("=========================")
    print("|| Fin de la selección ||")
    print("=========================")
    print("Fórmula obtenida:", f_actual)
    return f_actual




import numpy as np
from pandas import Series
from scipy.stats import f as fisher
from scipy.stats import ttest_1samp, ttest_ind, ttest_rel

from statsmodels.api import OLS, GLM
from statsmodels.genmod.families.family import Gaussian


class TestResults:
    def __init__(self, res: dict, names: dict):
        self.__dict__.update(res)
        self.__names = names

    def __repr__(self):
        # p-value
        p_value = self.p_value
        if p_value < 0.0001:
            p_value = "<0.0001"
        else:
            p_value = f"{p_value:0.4f}"
        # df
        df = self.df
        if isinstance(df, (float, int)):
            if df == float(int(df)):
                df = int(df)
        elif isinstance(df, dict):
            df = {k: f"{v:0.2f}" for k, v in df.items()}
        elif isinstance(df, float):
            df = f"{df:0.2f}"
        # estimates
        estimate = self.estimate
        if isinstance(estimate, list):
            estimate = [float(f"{e:0.6f}") for e in estimate]
        elif estimate == float(int(estimate)):
            estimate = int(estimate)
        elif isinstance(estimate, float):
            estimate = f"{estimate:0.6f}"

        string = f"""
    {self.method}
    {len(self.method) * "-"}
    {self.__names['statistic']} = {self.statistic:0.4f} | df: {df} | p-value = {p_value}
    alternative hypothesis: {self.__names["alternative"]}"""
        if self.__dict__.get("conf_int"):
            cl = self.conf_level * 100
            if cl == float(int(cl)):
                cl = int(cl)
            string += f"""
    {cl} percent confidence interval:
    {" "}{self.conf_int[0]:0.6f} {self.conf_int[1]:0.6f}"""
        string += f"""
    sample estimates:
    {" " * 2}{self.__names["estimate"]}: {estimate}
    """
        return string


def __var_test(x, y, ratio, alternative, conf_level):
    if not (isinstance(ratio, float) or isinstance(ratio, int)):
        raise ValueError("'ratio' must be a single positive number")
    else:
        if not ratio > 0:
            raise ValueError("'ratio' must be a single positive number")

    if alternative not in ["two-sided", "less", "greater"]:
        raise ValueError(
            "'alternative' must be one of 'two.sided', 'less', or 'greater'"
        )

    if not (isinstance(conf_level, float)):
        raise ValueError("'conf_level' must be a single number between 0 and 1")
    else:
        if not (conf_level > 0 and conf_level < 1):
            raise ValueError("'conf_level' must be a single number between 0 and 1")

    if isinstance(x, np.ndarray) or isinstance(x, Series):
        df_x = x.shape[0] - 1
        v_x = np.var(x)
    else:
        df_x = len(x) - 1
        v_x = np.var(x)

    if isinstance(y, np.ndarray) or isinstance(y, Series):
        df_y = y.shape[0] - 1
        v_y = np.var(y)
    else:
        df_y = len(y) - 1
        v_y = np.var(y)

    est = v_x / v_y
    stat = est / ratio

    f = fisher(dfn=df_x, dfd=df_y)
    pval = f.cdf(stat)

    if alternative == "two-sided":
        pval = 2 * min([pval, 1 - pval])
        beta = (1 - conf_level) / 2
        ci = [est / f.ppf(1 - beta), est / f.ppf(beta)]
    elif alternative == "greater":
        pval = 1 - pval
        ci = [est / f.ppf(conf_level), np.inf]
    else:
        ci = [0, est / f.ppf(1 - conf_level)]

    alternative_name = (
        "not equal to" if alternative == "two-sided" else f"{alternative} than"
    )

    names = {
        "statistic": "F",
        "estimate": "ratio of variances",
        "alternative": f"true ratio of variances is {alternative_name} {ratio}",
    }

    res = {
        "method": "F test to compare two variances",
        "statistic": stat,
        "estimate": est,
        "df": {"x": df_x, "y": df_y},
        "p_value": pval,
        "conf_int": ci,
        "conf_level": conf_level,
        "alternative": alternative,
    }

    return TestResults(res, names)


def __t_test(x, y, alternative, mu, paired, var_equal, conf_level):
    if isinstance(conf_level, float):
        if conf_level <= 0 or conf_level >= 1:
            raise ValueError("'conf_level' must be between 0 and 1")
    else:
        raise TypeError("'conf_level' must be a single number between 0 and 1")

    if isinstance(alternative, str):
        if alternative not in ["two-sided", "less", "greater"]:
            raise ValueError(
                "'alternative' must be one of 'two-sided', 'less', or 'greater'"
            )
    else:
        raise TypeError("'alternative' must be a string")

    if paired:
        if y is None:
            raise ValueError("If 'paired' is True, 'y' must be provided")
        else:
            result = ttest_rel(x, y + mu, alternative=alternative)
    else:
        if y is None:
            result = ttest_1samp(x, mu, alternative=alternative)
        else:
            result = ttest_ind(x, y + mu, alternative=alternative, equal_var=var_equal)

    lower, upper = result.confidence_interval(conf_level)
    ci = (lower, upper)
    t, p = result.statistic, result.pvalue
    df = result.df

    alternative_dire = (
        "not equal to" if alternative == "two-sided" else f"{alternative} than"
    )
    alternative_conc = "mean difference" if paired else "difference in means"

    if y is None:
        est = x.mean()
        estimate = "mean of x"
        alternative_name = f"true mean is {alternative_dire} {mu}"
        method = "One Sample t-test"
    else:
        welch = "Welch's " if (not var_equal and not paired) else ""
        title = "Paired" if paired else "Two Sample"
        est = [x.mean(), y.mean()]
        estimate = "[mean of x, mean of y]"
        alternative_name = f"true {alternative_conc} is {alternative_dire} {mu}"
        method = f"{welch}{title} t-test"

    names = {
        "statistic": "T",
        "estimate": estimate,
        "alternative": alternative_name,
    }

    res = {
        "method": method,
        "statistic": t,
        "estimate": est,
        "df": df,
        "p_value": p,
        "conf_int": ci,
        "mu": mu,
        "conf_level": conf_level,
        "alternative": alternative,
    }

    return TestResults(res, names)


def __deviance(fitted_model, response):
    if isinstance(fitted_model.model, OLS):
        return Gaussian().deviance(endog=response, mu=fitted_model.fittedvalues)
    elif isinstance(fitted_model.model, GLM):
        return fitted_model.deviance
    return sum(fitted_model.resid_dev ** 2) 


def __nested_models_test(fitted_small_model, fitted_big_model):
    response = fitted_big_model.model.data.endog
    n = fitted_big_model.fittedvalues.shape[0]
    p_big = fitted_big_model.params.shape[0] - 1
    p_small = fitted_small_model.params.shape[0] - 1
    d_big = __deviance(fitted_big_model, response)
    d_small = __deviance(fitted_small_model, response)
    df_num = p_big - p_small
    df_den = n - p_big - 1
    f_num = (d_small - d_big) / df_num
    f_den = d_big / df_den
    f_stat = f_num / f_den
    p_value = fisher(dfn=df_num, dfd=df_den).sf(f_stat)

    names = {
        "statistic": "F",
        "estimate": "Difference in deviances between models",
        "alternative": "big model is true",
    }

    res = {
        "method": "Nested models F-test",
        "statistic": f_stat,
        "estimate": d_small - d_big,
        "df": {"df_num": df_num, "df_den": df_den},
        "p_value": p_value,
    }

    return TestResults(res, names)
